from PyQt5.QtCore import QItemSelection, pyqtSignal
from PyQt5 import QtCore
from PyQt5.QtWidgets import QTreeView, QWidget, QAction, QFileDialog
from PyQt5.QtGui import QStandardItem
from silicon_analyser.helper.abstract.abstractmywindow import AbstractMyWindow
from silicon_analyser.grid import Grid
from silicon_analyser.helper.abstract.abstracttreehelper import AbstractTreeHelper
from silicon_analyser.helper.abstract.abstractimage import AbstractImage
from silicon_analyser.treeitem import TreeItem
from silicon_analyser.grid import Grid

class Tree(AbstractTreeHelper):
    _myWindow: AbstractMyWindow
    _actionGridAddRowTop: QAction
    _actionSaveModel: QAction
    _actionLoadModel: QAction
    _actionRemoveGrid: QAction
    _actionRemoveLabel: QAction
    _actionSaveAsCsv: QAction
    evtTreeSelectionChanged: pyqtSignal = pyqtSignal(QItemSelection)

    def __init__(self, parent: QWidget | None = ...) -> None:
        super().__init__(parent)
    
    def initialize(self, myWindow: AbstractMyWindow):
        self._myWindow = myWindow
        self.selectionModel().selectionChanged.connect(self.treeSelectionChanged)
        self._actionGridAddRowTop = self._myWindow._actionGridAddRowTop
        self._actionSaveModel = self._myWindow._actionSaveModel
        self._actionLoadModel = self._myWindow._actionLoadModel
        self._actionRemoveGrid = self._myWindow._actionRemoveGrid
        self._actionRemoveLabel = self._myWindow._actionRemoveLabel
        self._actionSaveAsCsv = self._myWindow._actionSaveAsCsv
        self.addAction(self._actionGridAddRowTop)
        self.addAction(self._actionSaveModel)
        self.addAction(self._actionLoadModel)
        self.addAction(self._actionRemoveGrid)
        self.addAction(self._actionRemoveLabel)
        self.addAction(self._actionSaveAsCsv)
        self._actionGridAddRowTop.triggered.connect(self.addTopRow)
        self._actionSaveModel.triggered.connect(self.saveModel)
        self._actionLoadModel.triggered.connect(self.loadModel)
        self._actionRemoveGrid.triggered.connect(self.removeGrid)
        self._actionRemoveLabel.triggered.connect(self.removeLabel)
        self._actionSaveAsCsv.triggered.connect(self.saveAsCsv)
    
    def saveAsCsv(self, *args, **kwargs):
        print("saveAsCsv")
        dlg = QFileDialog()
        dlg.setLabelText(QFileDialog.DialogLabel.Accept, "Save")
        filenames = []
        filenames = dlg.getSaveFileName(caption="Save csv",filter="Comma separated values (*.csv)",initialFilter="Comma separated values (*.csv)")
        if(len(filenames) >= 1):
            grid:Grid 
            if self.selectedType() == TreeItem.TYPE_GRID:
                grid = self.getSelectedGrid()
            if self.selectedType() == TreeItem.TYPE_AI_GRID:
                grid = self.getSelectedAIGrid()
            with open(filenames[0],"w") as f:
                for r in range(0,grid.rows):
                    cells = []
                    for c in range(0,grid.cols):
                        for l in list(grid.getLabels()):
                            if grid.isRectSet(c,r,l):
                                if l.isnumeric():
                                    cells.append(f'{l}')
                                else:
                                    cells.append(f'"{l}"')
                    f.write(",".join(cells))
                    f.write("\n")
    
    def saveModel(self, *args, **kwargs):
        print("saveModel")
        dlg = QFileDialog()
        dlg.setLabelText(QFileDialog.DialogLabel.Accept, "Save")
        filenames = []
        filenames = dlg.getSaveFileName(caption="Save trained model",filter="Trained model (*.h5)",initialFilter="Trained model (*.h5)")
        if(len(filenames) >= 1):
            from keras.models import Sequential
            grid = self.getSelectedGrid()
            model: Sequential = self._myWindow.getModel(grid.name)
            model.save(filenames[0])
    
    
    def recreateAiTree(self, grid:Grid):
        self.clearAIItem(grid.name)
        aiGrid, aiTreeItem = self.addTreeItem(grid.name,TreeItem.TYPE_AI_GRID)
        for l in list(grid.getLabels()):
            self.addTreeItem(l,TreeItem.TYPE_AI_GRID_ITEM,aiGrid,aiTreeItem)
        return aiGrid
    
    def loadModel(self, *args, **kwargs):
        print("loadModel")
        myWindow: AbstractMyWindow = self._myWindow
        dlg = QFileDialog()
        dlg.setFileMode(QFileDialog.ExistingFile)
        dlg.setNameFilter("Trained model (*.h5)")
        filenames = []
        if dlg.exec_():
            filenames = dlg.selectedFiles()
        if(len(filenames) == 1):
            from keras.models import load_model
            from silicon_analyser.helper.ai import appendFoundCellRects
            grid: Grid = self.getSelectedGrid()
            model = load_model(filenames[0])
            aiGrid = self.recreateAiTree(grid)
            myWindow.setLastModel(grid.name, model)
            img: AbstractImage = myWindow.getImage()
            appendFoundCellRects(img, grid, aiGrid, None, None, model)
            img.drawImage()
    
    def removeGrid(self, *args, **kwargs):
        print("removeGrid")
        myWindow: AbstractMyWindow = self._myWindow
        grid: Grid = self.getSelectedGrid()
        myWindow.getImage().removeGrid(grid.name)
        myWindow.getImage().drawImage()
        self.removeSelectedRow()

    def removeSelectedRow(self):
        index_list: list[QtCore.QModelIndex] = []                                                          
        for model_index in self.selectionModel().selectedRows():       
            index = QtCore.QPersistentModelIndex(model_index)         
            index_list.append(index)                                             

        for index in index_list:                                      
            self.model().removeRow(index.row(),index.parent())

    def removeLabel(self, *args, **kwargs):
        print("removeLabel")
        myWindow: AbstractMyWindow = self._myWindow
        label = self.selectedLabel()
        if(self.selectedType() == TreeItem.TYPE_GRID_ITEM):
            grid: Grid = self.getSelectedGrid()
            grid.removeRectGroup(label)
        elif(self.selectedType() == TreeItem.TYPE_MANUAL):
            myWindow.getImage().removeRectGroup(label)
        myWindow.getImage().drawImage()
        self.removeSelectedRow()

    def addTopRow(self, *args, **kwargs):
        print("addTopRow")
        grid: Grid = self.getSelectedGrid()
        grid.addTopRow()
        
    def treeSelectionChanged(self, selection: QItemSelection):
        print("treeSelectionChanged")
        myWindow: AbstractMyWindow = self._myWindow
        myWindow.reloadPropertyWindow(selection)
        tree = self
        selectedType = tree.selectedType()
        if((selectedType == TreeItem.TYPE_GRID_ITEM) or (selectedType == TreeItem.TYPE_GRID)):
            self._actionGridAddRowTop.setVisible(True)
            grid: Grid = self.getSelectedGrid()
            if myWindow.hasModel(grid.name):
                self._actionSaveModel.setVisible(True)
            self._actionLoadModel.setVisible(True)
        else:
            self._actionGridAddRowTop.setVisible(False)
            self._actionSaveModel.setVisible(False)
            self._actionLoadModel.setVisible(False)
        if(selectedType == TreeItem.TYPE_GRID):
            self._actionRemoveGrid.setVisible(True)
            self._actionSaveAsCsv.setVisible(True)
        else:
            self._actionRemoveGrid.setVisible(False)
            self._actionSaveAsCsv.setVisible(False)
        if(selectedType == TreeItem.TYPE_AI_GRID):
            self._actionSaveAsCsv.setVisible(True)
        else:
            self._actionSaveAsCsv.setVisible(False)
        if(selectedType == TreeItem.TYPE_MANUAL):
            self._actionRemoveLabel.setVisible(True)
        else:
            self._actionRemoveLabel.setVisible(False)
        if(selectedType == TreeItem.TYPE_GRID_ITEM):
            self._actionRemoveLabel.setVisible(True)
        else:
            self._actionRemoveLabel.setVisible(False)
        if(selectedType == TreeItem.TYPE_GRID_ITEM):
            myWindow.getImage().drawImage()
        if(selectedType == TreeItem.TYPE_AI_GRID_ITEM):
            myWindow.getImage().drawImage()
        self.evtTreeSelectionChanged.emit(selection)
            
    def addTreeItem(self, text, type="auto", parent_obj=None, parent_item=None) -> tuple[Grid, TreeItem]:
        obj = None
        myWindow: AbstractMyWindow = self._myWindow
        img = myWindow.getImage()
        tree: Tree = self
        if type == "auto":
            if tree.selectedType() == TreeItem.TYPE_GRID:
                type = TreeItem.TYPE_GRID_ITEM
            elif tree.selectedType() == TreeItem.TYPE_GRID_ITEM:
                type = TreeItem.TYPE_GRID_ITEM
            else:
                type = TreeItem.TYPE_MANUAL
        if type == TreeItem.TYPE_AI_GRID:
            obj = img.appendAIGrid(text)
            img.activateAIGrid(text)
        if type == TreeItem.TYPE_GRID:
            obj = img.appendGrid(text)
            img.activateGrid(text)
        if type == TreeItem.TYPE_AI_GRID_ITEM:
            grid = parent_obj
            if grid is None:
                grid = self.getSelectedAIGrid()
            obj = img.appendAIGridRectGroup(grid,text)
            img.activateAIGridRectGroup(grid,text)
        if type == TreeItem.TYPE_GRID_ITEM:
            grid = parent_obj
            if grid is None:
                grid = self.getSelectedGrid()
            obj = img.appendGridRectGroup(grid,text)
            img.activateGridRectGroup(grid,text)
        if type == TreeItem.TYPE_MANUAL:
            img.appendRectGroup(text)
            img.activateRectGroup(text)
        if type == TreeItem.TYPE_AI:
            img.appendAIRectGroup(text)
            img.activateAIRectGroup(text)
        item: TreeItem = TreeItem(text, type, obj)
        if type == TreeItem.TYPE_MANUAL:
            treeItem: QStandardItem = myWindow.getManualItem()
        if type == TreeItem.TYPE_AI:
            treeItem: QStandardItem = myWindow.getAIItem()
        if type == TreeItem.TYPE_GRID:
            treeItem: QStandardItem = myWindow.getManualItem()
        if type == TreeItem.TYPE_AI_GRID:
            treeItem: QStandardItem = myWindow.getAIItem()
        if type == TreeItem.TYPE_GRID_ITEM:
            treeItem: QStandardItem = self.getSelectedItem().parent()
        if type == TreeItem.TYPE_AI_GRID_ITEM:
            treeItem: QStandardItem = self.getSelectedItem().parent()
        if parent_item is not None:
            treeItem: QStandardItem = parent_item
        tree: QTreeView = self
        treeItem.appendRow(item)
        tree.expandAll()
        selModel = tree.selectionModel()
        tree.clearSelection()
        selModel.select(tree.model().indexFromItem(item), selModel.Select)
        item.setFlags(QtCore.Qt.ItemIsUserCheckable |
                        QtCore.Qt.ItemIsEnabled |
                        QtCore.Qt.ItemIsSelectable)
        item.setCheckState(QtCore.Qt.Checked)
        if type in [TreeItem.TYPE_AI_GRID,TreeItem.TYPE_GRID,TreeItem.TYPE_MANUAL]:
            item.onChecked = self.itemChecked
        if type == TreeItem.TYPE_GRID_ITEM:
            item.onChecked = self.gridRectGroupChecked
        if type == TreeItem.TYPE_AI_GRID_ITEM:
            item.onChecked = self.aiGridRectGroupChecked
        if type == TreeItem.TYPE_AI:
            item.onChecked = self.aiItemChecked
        return obj, item
    
    def getSelectedItem(self) -> QStandardItem:
        tree: QTreeView = self
        selection = tree.selectedIndexes()
        cnt = len(selection)
        if(cnt == 0):
            return
        sel = selection[0]
        return tree.model().itemFromIndex(sel)
    
    def getSelectedGrid(self):
        myWindow: AbstractMyWindow = self._myWindow
        type = self.getSelectedItem().data(TreeItem.TYPE)
        if type == TreeItem.TYPE_GRID_ITEM:
            return self.getSelectedItem().parent().data(TreeItem.OBJECT)
        if type == TreeItem.TYPE_AI_GRID_ITEM:
            gridName = self.getSelectedItem().parent().data(TreeItem.TEXT)
            manual: QStandardItem = myWindow.getManualItem()
            for r in range(0,manual.rowCount()):
                if manual.child(r).data(TreeItem.TEXT) == gridName:
                    return manual.child(r).data(TreeItem.OBJECT)
        if type == TreeItem.TYPE_AI_GRID:
            gridName = self.getSelectedItem().data(TreeItem.TEXT)
            manual: QStandardItem = myWindow.getManualItem()
            for r in range(0,manual.rowCount()):
                if manual.child(r).data(TreeItem.TEXT) == gridName:
                    return manual.child(r).data(TreeItem.OBJECT)
        return None if type != TreeItem.TYPE_GRID else self.getSelectedItem().data(TreeItem.OBJECT)

    def getSelectedAIGrid(self):
        type = self.getSelectedItem().data(TreeItem.TYPE)
        return None if type != TreeItem.TYPE_AI_GRID else self.getSelectedItem().data(TreeItem.OBJECT)
    
    def isItemSelected(self, rectLabel, gridName, gridItemType) -> bool:
        type = self.getSelectedItem().data(TreeItem.TYPE)
        if type != gridItemType:
            return False
        if self.getSelectedItem().parent().data(TreeItem.TEXT) != gridName:
            return False
        if self.getSelectedItem().data(TreeItem.TEXT) != rectLabel:
            return False
        return True
    
    def clearAIItem(self, name:str = None):
        myWindow: AbstractMyWindow = self._myWindow
        treeAIItem = myWindow.getAIItem()
        if treeAIItem.hasChildren():
            if name is None:
                treeAIItem.removeRows(0,treeAIItem.rowCount())
                treeAIItem.clearData()
                myWindow.getImage().clearAIRects()
            else:
                for i in range(0,treeAIItem.rowCount):
                    child = treeAIItem.child(i)
                    if child.data(TreeItem.TEXT) == name:
                        child.removeRows(0,treeAIItem.rowCount())
                        child.clearData()
    
    def selectedType(self) -> str:
        sel = self.getSelectedItem()
        if sel is None:
            return None
        return sel.data(TreeItem.TYPE)
    
    def selectedLabel(self) -> str:
        sel = self.getSelectedItem()
        if sel is None:
            return None
        return sel.data(TreeItem.TEXT)
    
    def aiGridRectGroupChecked(self, treeItem: TreeItem, checked, text):
        myWindow: AbstractMyWindow = self._myWindow
        img = myWindow.getImage()
        grid = treeItem.parent().data(TreeItem.OBJECT)
        if checked:
            img.activateAIGridRectGroup(grid,text)
        else:
            img.deactivateAIGridRectGroup(grid,text)
        img.drawImage()
        
    def gridRectGroupChecked(self, treeItem: TreeItem, checked, text):
        myWindow: AbstractMyWindow = self._myWindow
        img = myWindow.getImage()
        grid = treeItem.parent().data(TreeItem.OBJECT)
        if checked:
            img.activateGridRectGroup(grid,text)
        else:
            img.deactivateGridRectGroup(grid,text)
        img.drawImage()
        
    def aiItemChecked(self, treeItem: TreeItem, checked, text):
        myWindow: AbstractMyWindow = self._myWindow
        img = myWindow.getImage()
        if checked:
            img.activateAIRectGroup(text)
        else:
            img.deactivateAIRectGroup(text)
    
    def itemChecked(self, treeItem: TreeItem, checked, text):
        myWindow: AbstractMyWindow = self._myWindow
        img = myWindow.getImage()
        if checked:
            img.activateRectGroup(text)
        else:
            img.deactivateRectGroup(text)


from PyQt5.QtWidgets import QInputDialog, QPushButton
from PyQt5.QtGui import QMouseEvent
from silicon_analyser.helper.abstract.abstractmywindow import AbstractMyWindow
from silicon_analyser.treeitem import TreeItem

class AddGridBtn(QPushButton):
    def __init__(self, text: str):
        QPushButton.__init__(self, text)
        self.clicked.connect(self.buttonClicked)
        self._gridIdx = 0

    def buttonClicked(self, event: QMouseEvent):
        print("AddGridBtn: buttonClicked")
        self._myWindow.getTree().addTreeItem("grid_%d" % self._gridIdx,TreeItem.TYPE_GRID)
        self._gridIdx += 1
    
    def initialize(self, myWindow: AbstractMyWindow):
        self._myWindow = myWindow


from sklearn.model_selection import train_test_split
import os
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtCore import QObject, QThread, pyqtSignal
from PyQt5.QtGui import QMouseEvent
import random
import numpy as np
import sys
from silicon_analyser.dialogs.compute_stats import ComputeStatsDlg
from silicon_analyser.helper.abstract.abstractmywindow import AbstractMyWindow
from silicon_analyser.grid import Grid
from silicon_analyser.treeitem import TreeItem
from silicon_analyser.helper.ai import appendFoundCellRects
# because debugging in multithread can cause problems (breakpoints don't work)
# I added this method ...
def isSingleThread():
    if "SINGLE_THREAD" not in os.environ:
        return False
    return "1" == os.environ["SINGLE_THREAD"]

def blockshaped(arr, nrows, ncols):
    """
    Return an array of shape (n, nrows, ncols) where
    n * nrows * ncols = arr.size

    If arr is a 2D array, the returned array should look like n subblocks with
    each subblock preserving the "physical" layout of arr.
    """
    h, w, v = arr.shape
    assert h % nrows == 0, f"{h} rows is not evenly divisible by {nrows}"
    assert w % ncols == 0, f"{w} cols is not evenly divisible by {ncols}"
    return (arr.reshape(h//nrows, nrows, -1, ncols, v)
               .swapaxes(1,2)
               .reshape(-1, nrows, ncols, v))

class Worker(QObject):
    _computeStatsDlg: ComputeStatsDlg
    finished = pyqtSignal()
    progress = pyqtSignal(int)
    
    def __init__(self,myWindow: AbstractMyWindow, computeStatsDlg: ComputeStatsDlg):
        QObject.__init__(self)
        self._myWindow = myWindow
        self._computeStatsDlg = computeStatsDlg

    def run(self):
        self._computeStatsDlg.setStatus("Initialize keras ...")
        os.environ["KERAS_BACKEND"] = "torch"
        import keras
        self._computeStatsDlg.setStatus("... finished Initializing keras")
        
        def appendFoundRects(self, rectKeys, maxW, maxH, model_conv, MP):
            import keras_cv
            bx = 10000
            by = 10000
            data = self._myWindow.getImage().fetchFullData()[:,:,0:3]
            dx = data.shape[0]
            dy = data.shape[1]
            dx -= dx % bx
            dy -= dy % by
            maxCnt = 20
            data = data[0:dx,0:dy]
            print("data.shape", data.shape)
            model_conv.build(input_shape=[1,*data.shape[0:2],3])
            subdataIdx = 0
            cnt = 0
            for subdata in blockshaped(data, bx, by):
                if cnt < maxCnt:
                    print("subdata.shape",subdata.shape)
                    r = model_conv(np.array([subdata]))[0]
                    r = r.cpu().detach().numpy()
                    print("r.shape",r.shape)
                    for rx in range(0,r.shape[0]):
                        for ry in range(0,r.shape[1]):
                            if cnt >= maxCnt:
                                continue
                            x = rx + int(subdataIdx * subdata.shape[0]) % data.shape[0]
                            y = ry + (int(subdataIdx * subdata.shape[0]) // data.shape[0])*subdata.shape[1]
                            idx = np.argmax(r[rx,ry])
                            if(idx > 0):
                                x *= MP
                                y *= MP
                                key = rectKeys[idx-1]
                                ex = x+maxW
                                ey = y+maxH
                                self._myWindow.getImage().appendAIRect(key,x,y,ex,ey)
                                cnt += 1
                subdataIdx += 1

        class MyPlottingCallback(keras.callbacks.Callback):
            def __init__(self, plotDlg: ComputeStatsDlg):
                self._plotDlg = plotDlg
            
            def on_epoch_end(self, epoch, logs=None): 
                self._plotDlg.request_graph_update.emit(logs, epoch)
                if self._plotDlg.isStop():
                    self.model.stop_training = True
            
        class MyThresholdCallback(keras.callbacks.Callback):
            def __init__(self, threshold, patience):
                super(MyThresholdCallback, self).__init__()
                self.threshold = threshold
                self.patience = patience
                self.count = 0

            def on_epoch_end(self, epoch, logs=None): 
                val_acc = logs["val_accuracy"]
                if val_acc >= self.threshold:
                    if self.count > self.patience:
                        self.model.stop_training = True
                    else:
                        self.count += 1
                else:
                    self.count = 0
        
        print("_worker runs")
        try:
            self._computeStatsDlg.reset()
            selectedType = self._myWindow.getTree().selectedType()
            gridMode = (selectedType == TreeItem.TYPE_GRID) or (selectedType == TreeItem.TYPE_GRID_ITEM) or (selectedType == TreeItem.TYPE_AI_GRID) or (selectedType == TreeItem.TYPE_AI_GRID_ITEM)
            rects = self._myWindow.getImage().getRects()
            ignoreRects = self._myWindow.getImage().getAIIgnoreRects()
            self._myWindow.setStatusText("training in progress")

            grid = None        
            if gridMode:
                grid: Grid = self._myWindow.getTree().getSelectedGrid()
                maxW, maxH, labels, countGroups, MP, train, vals = self.initTrainAndValsForGrid(grid)
            else:
                maxW, maxH, countGroups, MP, train, vals = self.initTrainAndValsForRects(rects, ignoreRects)

            model_conv, model_train = self.createModels(keras, countGroups, maxW, maxH, MP)
            
            if gridMode:
                aiGrid = self.prepareGridData(grid, labels, countGroups, maxW, maxH, train, vals)
            else:
                self.prepareRectData(rects, countGroups, ignoreRects, maxW, maxH, train, vals)
            
            callbacks = [MyPlottingCallback(self._computeStatsDlg),MyThresholdCallback(0.99,200)]
            model_train.compile(optimizer=keras.optimizers.Adam(learning_rate=0.002),loss="binary_crossentropy",metrics=["accuracy"])
            print("before fit")
            print("vals",vals)
            print("train.shape",train.shape)
            print("vals.shape",vals.shape)
            
            xtraining, ytraining, xvalidation, yvalidation = train_test_split(train,vals,test_size=0.1) 

            history = model_train.fit(
                x = xtraining,
                y = xvalidation,
                epochs = 100000,
                verbose = 0,
                validation_data=(ytraining,yvalidation),
                shuffle = True,
                batch_size = 512,
                callbacks = callbacks
            )
            
            print("after fit")
            self._myWindow.setStatusText("Accuracy: %s" % history.history["accuracy"][-1])
            
            if gridMode:
                appendFoundCellRects(self._myWindow.getImage(), grid, aiGrid, maxW, maxH, model_train)
            else:
                appendFoundRects(self, list(rects.keys()), maxW, maxH, model_conv, MP)
            
            self._myWindow.getImage().drawImage()
            self._myWindow.setLastModel(aiGrid.name, model_train)
        except Exception as e:
            self._computeStatsDlg.setError(e)
        finally:
            self.finished.emit()
        
    def initTrainAndValsForRects(self, rects, ignoreRects):
        maxW = 0
        maxH = 0
        cntTotal = 0
        for k in rects.keys():
            for x,y,ex,ey in rects[k]:
                maxW = max(maxW, ex-x)
                maxH = max(maxH, ey-y)
                cntTotal += 1

        for x,y,ex,ey in ignoreRects:
            maxW = max(maxW, ex-x)
            maxH = max(maxH, ey-y)
            cntTotal += 1
        cntTotal += 100
        MP=5
        if maxW % MP != 0:
            maxW += MP - (maxW % MP)
        if maxH % MP != 0:
            maxH += MP - (maxH % MP)
        countGroups = len(rects.keys())+1
        train = np.zeros(shape=(cntTotal,maxH,maxW,3),dtype=np.float32)
        vals = np.zeros(shape=(cntTotal,countGroups),dtype=np.float32)
        print("train.shape:",train.shape)
        print("vals.shape:",vals.shape)
        return maxW, maxH, countGroups, MP, train, vals

    def initTrainAndValsForGrid(self, grid:Grid):
        labels = grid.getLabels()
        countGroups = len(labels)
        cntTotal = 0
        for l in labels:
            for cx in range(0,grid.cols):
                for cy in range(0,grid.rows):
                    if grid.isRectSet(cx,cy,l):
                        cntTotal += 1
        maxW = grid.width//grid.cols
        maxH = grid.height//grid.rows
        MP = 5
        if maxW % MP != 0:
            maxW += MP - (maxW % MP)
        if maxH % MP != 0:
            maxH += MP - (maxH % MP)
        train = np.zeros(shape=(cntTotal,maxH,maxW,3),dtype=np.float32)
        vals = np.zeros(shape=(cntTotal,countGroups),dtype=np.float32)
        print("train.shape:",train.shape)
        print("vals.shape:",vals.shape)
        return maxW, maxH, labels, countGroups, MP, train, vals

    def prepareRectData(self, rects, countGroups, ignoreRects, maxW, maxH, train, vals):
        self._myWindow.getTree().clearAIItem()
        i = 0
        ii = 1
        print(f"adding ignore values ({len(ignoreRects)})")
            
        for n in range(0,100):
            x = random.randint(0,self._myWindow.imageWidth()-maxW)
            y = random.randint(0,self._myWindow.imageHeight()-maxH)
            data = self._myWindow.getImage().fetchData(x,y,x+maxW-1,y+maxH-1)
            train[i,0:data.shape[0],0:data.shape[1]] = data
            vals[i] = np.eye(countGroups,dtype=np.float32)[0]
            i += 1
            
        for x,y,ex,ey in ignoreRects:
            data = self._myWindow.getImage().fetchData(x,y,x+maxW-1,y+maxH-1)
            train[i,0:data.shape[0],0:data.shape[1]] = data
            vals[i] = np.eye(countGroups,dtype=np.float32)[0]
            i += 1
            
        print("adding values")
        for k in rects.keys():
            self._myWindow.getTree().addTreeItem(k,"ai")
            for x,y,ex,ey in rects[k]:
                data = self._myWindow.getImage().fetchData(x,y,x+maxW-1,y+maxH-1)
                    #np.save(f"samples/{i}.bin",data)
                train[i,0:maxH,0:maxW] = data
                vals[i] = np.eye(countGroups,dtype=np.float32)[ii]
                i += 1
            ii += 1

    def prepareGridData(self, grid: Grid, labels, countGroups, maxW, maxH, train, vals) -> Grid:
        self._myWindow.getTree().clearAIItem()
        i = 0
        ii = 0
        aiGrid, aiTreeItem = self._myWindow.getTree().addTreeItem(grid.name,TreeItem.TYPE_AI_GRID)
        for l in labels:
            self._myWindow.getTree().addTreeItem(l,TreeItem.TYPE_AI_GRID_ITEM,aiGrid,aiTreeItem)
            for cx in range(0,grid.cols):
                for cy in range(0,grid.rows):
                    if grid.isRectSet(cx,cy,l):
                        x = int(grid.x + cx*maxW)
                        y = int(grid.y + cy*maxH)
                        ex = x + maxW - 1
                        ey = y + maxH - 1
                        data = self._myWindow.getImage().fetchData(x,y,ex,ey)
                        train[i,0:maxH,0:maxW] = data
                        #np.save(f"samples/{l}_{cx:03d}_{cy:03d}_{x:04d}_{y:04d}.bin",data)
                        vals[i] = np.eye(countGroups,dtype=np.float32)[ii]
                        i += 1
            ii += 1
        return aiGrid

    def createModels(self, keras, countGroups, maxW, maxH, MP):
        conv2 = keras.layers.Conv2D(
                countGroups*5,
                name="conv2d_2",
                kernel_size=(int(maxH//MP-5),int(maxW//MP-5)),
                activation="softmax"
        )

        conv3 = keras.layers.Conv2D(
                countGroups,
                name="conv2d_3",
                kernel_size=(1+5,1+5),
                activation="softmax"
        )

        model_conv = keras.Sequential([
            keras.layers.BatchNormalization(),
            keras.layers.MaxPooling2D(MP),
            conv2,
            conv3
        ])

        model_train = keras.Sequential([
            keras.layers.Input((maxH,maxW,3)),
            keras.layers.BatchNormalization(),
            keras.layers.MaxPooling2D(MP),
            conv2,
            conv3,
            keras.layers.Reshape((countGroups,))
        ])
        model_train.summary()
        return model_conv,model_train
        
class ComputeBtn(QPushButton):
    _computeStatsDlg: ComputeStatsDlg

    def __init__(self, text: str):
        QPushButton.__init__(self, text)
        self._computeStatsDlg = ComputeStatsDlg()
        self.clicked.connect(self.buttonClicked)
      
    def initialize(self, myWindow: AbstractMyWindow):
        self._myWindow = myWindow
    
    def buttonClicked(self, event: QMouseEvent):
        print("ComputeBtn: buttonClicked")
        self._computeStatsDlg.show()
        self._worker = Worker(self._myWindow,self._computeStatsDlg)
        self._worker.finished.connect(self._worker.deleteLater)
        self._thread = QThread()
        self._worker.finished.connect(self._thread.quit)
        self._worker.moveToThread(self._thread)
        self._thread.started.connect(self._worker.run)
        self._thread.finished.connect(self._thread.deleteLater)
        self._thread.finished.connect(
            lambda: print("finished")
        )
        self._thread.finished.connect(
            lambda: self.setText("Compute")
        )
        self._thread.finished.connect(
            lambda: self.setEnabled(True)
        )
        self.setText("...")
        self.setEnabled(False)
        if isSingleThread():
            self._worker.run()
        else:
            self._thread.start()


from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QStandardItem
from PyQt5.QtWidgets import QTreeView

from silicon_analyser.grid import Grid
from silicon_analyser.treeitem import TreeItem

class AbstractTreeHelper(QTreeView):
    evtTreeSelectionChanged: pyqtSignal
    
    def selectedType(self) -> str:
        raise NotImplementedError()
    
    def selectedLabel(self) -> str:
        raise NotImplementedError()

    def getSelectedItem(self) -> QStandardItem:
        raise NotImplementedError()
    
    def getSelectedGrid(self):
       raise NotImplementedError()

    def getSelectedAIGrid(self):
        raise NotImplementedError()
    
    def isItemSelected(self, rectLabel, gridName, gridItemType) -> bool:
        raise NotImplementedError()
            
    def addTreeItem(self, text, type="auto", parent_obj=None, parent_item=None) -> tuple[Grid, TreeItem]:
        raise NotImplementedError()
    
    def clearAIItem(self):
        raise NotImplementedError()


from PyQt5.QtWidgets import QTableView
from PyQt5.QtCore import QItemSelection, QModelIndex, Qt
from PyQt5.QtGui import QStandardItem, QStandardItemModel
from silicon_analyser.savefiles import saveGrids
from silicon_analyser.grid import Grid
from silicon_analyser.treeitem import TreeItem
from silicon_analyser.helper.abstract.abstractmywindow import AbstractMyWindow

class PropertiesUtil:
    def __init__(self, myWindow: AbstractMyWindow,properties: QTableView):
        self._properties = properties
        self._myWindow = myWindow
        properties.model().dataChanged.connect(self.dataChanged)
    
    def dataChanged(self,*args,**kwargs):
        print("dataChanged")
        valueCol: QModelIndex = args[0]
        if valueCol.column() != 1:
            return
        row = valueCol.row()
        model: QStandardItemModel = self._properties.model()
        nameCol = model.item(row,0)
        name = nameCol.data(Qt.DisplayRole)
        value = valueCol.data(Qt.DisplayRole)
        print(name)
        print(value)
        grid: Grid = self._myWindow.getTree().getSelectedGrid()
        if(name == "cols"):
            grid.cols = int(value)
        if(name == "rows"):
            grid.rows = int(value)
        if(name == "x"):
            grid.x = int(value)
        if(name == "y"):
            grid.y = int(value)
        if(name == "width"):
            grid.width = int(value)
        if(name == "height"):
            grid.height = int(value)
        self._myWindow.getImage().drawImage()
        saveGrids(self._myWindow.getImage().getGrids())
    
    def reloadPropertyWindow(self,selection: QItemSelection):
        table: QTableView = self._properties
        model = table.model()
        model.removeRows(0,model.rowCount())
        cnt = selection.count()
        if(cnt == 0):
            return
        sel = selection.takeFirst()
        if(len(sel.indexes())==0):
            return
        typeStr = sel.indexes()[0].data(TreeItem.TYPE)
        if(typeStr == TreeItem.TYPE_GRID):
            grid: Grid = sel.indexes()[0].data(TreeItem.OBJECT)
            self.reloadProperyWindowByGrid(grid)

    def reloadProperyWindowByGrid(self, grid):
        table: QTableView = self._properties
        model = table.model()
        rowPosition = model.rowCount()
        model.removeRows(0,model.rowCount())
        model.insertRow(rowPosition,[QStandardItem("name"),QStandardItem(grid.name)])
        if grid is not None:
            rowPosition = model.rowCount()
            model.insertRow(rowPosition,[QStandardItem("x"),QStandardItem(f"{grid.x}")])
            rowPosition = model.rowCount()
            model.insertRow(rowPosition,[QStandardItem("y"),QStandardItem(f"{grid.y}")])
            rowPosition = model.rowCount()
            model.insertRow(rowPosition,[QStandardItem("cols"),QStandardItem(f"{grid.cols}")])
            rowPosition = model.rowCount()
            model.insertRow(rowPosition,[QStandardItem("rows"),QStandardItem(f"{grid.rows}")])
            rowPosition = model.rowCount()
            model.insertRow(rowPosition,[QStandardItem("width"),QStandardItem(f"{grid.width}")])
            rowPosition = model.rowCount()
            model.insertRow(rowPosition,[QStandardItem("height"),QStandardItem(f"{grid.height}")])

  


from PyQt5.QtWidgets import QLabel
from PyQt5.QtCore import QSize, Qt
from PyQt5.QtGui import QPixmap, QPainter, QPen, QMouseEvent
from silicon_analyser.helper.abstract.abstractmywindow import AbstractMyWindow

class MiniMap(QLabel):
    def __init__(self, parent):
        QLabel.__init__(self, parent)
    
    def initialize(self, myWindow: AbstractMyWindow, pixmap: QPixmap):
        self._myWindow = myWindow
        self._pixmap = pixmap
        
    def mouseReleaseEvent(self,event: QMouseEvent):
        print("clicked")
        sc1x = 300/self._pixmap.width()
        sc1y = 300/self._pixmap.height()
        self._myWindow.setPosX(int(event.x()/sc1x))
        self._myWindow.setPosY(int(event.y()/sc1y))
        self._myWindow.drawImgAndMinimap()
        print(self._myWindow.getPosX(),self._myWindow.getPosY())

    def scaleMinimapX(self):
        return 300/self._pixmap.width()
    
    def scaleMinimapY(self):
        return 300/self._pixmap.height()

    def drawMinimap(self):
        minimapPm = self._pixmap.scaled(QSize(300, 300))
        posX, posY = self._myWindow.getPos()
        scale = self._myWindow.getScale()
        qp = QPainter(minimapPm)
        pen = QPen(Qt.red, 2)
        qp.setPen(pen)
        sc1x = self.scaleMinimapX()
        sc2x = sc1x/scale
        sc1y = self.scaleMinimapY()
        sc2y = sc1y/scale
        qp.drawRect(int(posX*sc1x),int(posY*sc1y),int(300*sc2x),int(300*sc2y))
        qp.end()
        self.setPixmap(minimapPm)


from PyQt5 import uic
from PyQt5 import QtGui
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QDialog, QLabel, QPushButton
from pyqtgraph import PlotWidget
from numpy import linspace
import os.path as p

class ComputeStatsDlg(QDialog):
    _lossgraph: PlotWidget
    _accuracygraph: PlotWidget
    _loss: any
    _val_loss: any
    _accuracy: any
    _val_accuracy: any
    _lblStatus: QLabel
    _btnStop: QPushButton
    request_graph_update = pyqtSignal([dict,int])
    
    def __init__(self):
        super().__init__()
        uic.loadUi(p.abspath(p.join(p.dirname(__file__), '.')) + '/compute_stats.ui', self)
        self._loss = linspace(0,0,self.width())
        self._val_loss = linspace(0,0,self.width())
        self._accuracy = linspace(0,0,self.width())
        self._val_accuracy = linspace(0,0,self.width())
        
        legend = self._lossgraph.addLegend()
        self._loss_curve = self._lossgraph.plot(pen='r')
        self._val_loss_curve = self._lossgraph.plot(pen='g')
        legend.addItem(self._loss_curve,'loss')
        legend.addItem(self._val_loss_curve ,'val_loss')
        
        legend = self._accuracygraph.addLegend()
        self._accuracy_curve = self._accuracygraph.plot(pen='r')
        self._val_accuracy_curve = self._accuracygraph.plot(pen='g')
        legend.addItem(self._accuracy_curve,'accuracy')
        legend.addItem(self._val_accuracy_curve,'val_accuracy')
        
        self._btnStop.clicked.connect(self.stop)
        self.request_graph_update.connect(self.updateGraph)
        
        self._lastX = 0
        self._stop = False
    
    def isStop(self):
        return self._stop

    def stop(self):
        self._stop = True
    
    def reset(self):
        self._stop = False
    
    def updateGraph(self,logs,epoch):
        self._lastX += 1

        self._loss[:-1] = self._loss[1:]
        self._val_loss[:-1] = self._val_loss[1:]
        self._accuracy[:-1] = self._accuracy[1:]
        self._val_accuracy[:-1] = self._val_accuracy[1:]

        loss, val_loss, accuracy, val_accuracy = logs["loss"], logs["val_loss"], logs["accuracy"], logs["val_accuracy"]

        self._loss[-1] = loss
        self._val_loss[-1] = val_loss
        self._accuracy[-1] = accuracy
        self._val_accuracy[-1] = val_accuracy
        
        self._loss_curve.setData(self._loss)
        self._loss_curve.setPos(self._lastX,0)
        self._val_loss_curve.setData(self._val_loss)
        self._val_loss_curve.setPos(self._lastX,0)

        self._accuracy_curve.setData(self._accuracy)
        self._accuracy_curve.setPos(self._lastX,0)
        self._val_accuracy_curve.setData(self._val_accuracy)
        self._val_accuracy_curve.setPos(self._lastX,0)
        
        self._lblStatus.setText(f"Status: epoch: {epoch: <10} loss: {loss:.6f} val_loss: {val_loss:.6f} acc: {accuracy:.2f} val_acc: {val_accuracy:.2f}")

        QtGui.QGuiApplication.processEvents()
    
    def setError(self,error: Exception):
        self._lblStatus.setStyleSheet("QLabel { background-color : red; color : black; }")
        self._lblStatus.setText(str(error))
        QtGui.QGuiApplication.processEvents()
    
    def setStatus(self,text: str):
        self._lblStatus.setStyleSheet("QLabel { color : black; }")
        self._lblStatus.setText(text)
        QtGui.QGuiApplication.processEvents()


from PyQt5.QtCore import Qt
from PyQt5.QtGui import QStandardItem
import typing

class TreeItem(QStandardItem):
    TYPE = 0x1337
    OBJECT = 0x1338
    TEXT = 0x1339
    
    TYPE_GRID = "grid"
    TYPE_AI_GRID = "aiGrid"
    TYPE_GRID_ITEM = "gridItem"
    TYPE_AI_GRID_ITEM = "aiGridItem"
    TYPE_MANUAL = "manual"
    TYPE_AI = "ai"
    
    def __init__(self, text, type, obj):
        QStandardItem.__init__(self,text)
        self.onChecked = None
        self._text = text
        self._type = type
        self._obj  = obj
    
    def data(self, role: int = ...):
        if(role == TreeItem.TYPE):
            return self._type
        if(role == TreeItem.OBJECT):
            return self._obj
        if(role == TreeItem.TEXT):
            return self._text
        return super().data(role)
    
    def getObject(self):
        return self._obj
    
    def setData(self, value: typing.Any, role: int = ...) -> None:
        if role == Qt.CheckStateRole:
            if self.onChecked is not None:
                if value == 0:
                    self.onChecked(self,False,self._text)
                else:
                    self.onChecked(self,True,self._text)
        return super().setData(value, role)
    
    


import numpy as np
import keras
from silicon_analyser.grid import Grid, getAllCellRects
from silicon_analyser.helper.abstract.abstractimage import AbstractImage

def getDefaultMaxWMaxH(grid: Grid):
    maxW = grid.width//grid.cols
    maxH = grid.height//grid.rows
    MP = 5
    if maxW % MP != 0:
        maxW += MP - (maxW % MP)
    if maxH % MP != 0:
        maxH += MP - (maxH % MP)
    return maxW, maxH

def appendFoundCellRects(img: AbstractImage, grid: Grid, aiGrid: Grid, maxW, maxH, model: keras.Sequential):
    if maxW is None or maxH is None:
        maxW, maxH = getDefaultMaxWMaxH(grid)
    labels = list(grid.getLabels())
    allCellRects = getAllCellRects(grid)
    dataList = []
    dataIndexes = []
    for cx,cy in allCellRects:
        x = int(grid.x + cx*maxW)
        y = int(grid.y + cy*maxH)
        ex = x + maxW - 1
        ey = y + maxH - 1
        dataList.append(img.fetchData(x,y,ex,ey))
        dataIndexes.append((cx,cy))
    data = np.array(dataList,dtype=np.float32)
    print("data.shape",data.shape)
    r = model.predict(data)
    print("r.shape",r.shape)
    for ri in range(0,r.shape[0]):
        currentRec = r[ri]
        lblIdx = np.argmax(currentRec)
        rx, ry = dataIndexes[ri]
        aiGrid.setRect(rx,ry,labels[lblIdx])


from PyQt5.QtWidgets import QApplication
import sys
from silicon_analyser.mywindow import MyWindow

def main_cli():
    app = QApplication(sys.argv)

    window = MyWindow()

    window.show()
    app.exec()

if __name__ == "__main__":
    main_cli()


from PyQt5.QtWidgets import QInputDialog, QPushButton
from PyQt5.QtGui import QMouseEvent
from silicon_analyser.helper.abstract.abstractmywindow import AbstractMyWindow

class AddLabelBtn(QPushButton):
    def __init__(self, text: str):
        QPushButton.__init__(self, text)
        self.clicked.connect(self.buttonClicked)

    def buttonClicked(self, event: QMouseEvent):
        print("AddLabelBtn: buttonClicked")
        text, ok = QInputDialog.getText(self, 'Label Input Dialog', 'Enter your label:')
        if ok:
            self._myWindow.getTree().addTreeItem(text)
            
    def initialize(self, myWindow: AbstractMyWindow):
        self._myWindow = myWindow


from PyQt5.QtWidgets import QLabel

from silicon_analyser.grid import Grid
from silicon_analyser.rect import Rect

class AbstractImage(QLabel):
    
    def clearAIRects(self):
        raise NotImplementedError()
    
    def fetchData(self,x,y,ex,ey):
        raise NotImplementedError()
    
    def drawImage(self):
        raise NotImplementedError()
    
    def fetchFullData(self):
        raise NotImplementedError()
    
    def appendRectGroup(self, text):
         raise NotImplementedError()

    def appendAIRectGroup(self, text):
         raise NotImplementedError()
    
    def appendGridRectGroup(self, grid, text):
        raise NotImplementedError()
    
    def appendAIGridRectGroup(self, grid, text):
        raise NotImplementedError()
    
    def activateGridRectGroup(self, grid, text):
        raise NotImplementedError()
    
    def deactivateGridRectGroup(self, grid, text):
        raise NotImplementedError()
    
    def activateAIGridRectGroup(self, grid, text):
        raise NotImplementedError()
    
    def deactivateAIGridRectGroup(self, grid, text):
        raise NotImplementedError()
    
    def activateRectGroup(self, text):
        raise NotImplementedError()

    def deactivateRectGroup(self, text):
        raise NotImplementedError()
    
    def activateAIRectGroup(self, text):
        raise NotImplementedError()

    def deactivateAIRectGroup(self, text):
        raise NotImplementedError()
    
    def appendAIRect(self,key,x,y,ex,ey):
        raise NotImplementedError()
    
    def appendAIGrid(self, text):
        raise NotImplementedError()
    
    def activateAIGrid(self, text):
        raise NotImplementedError()
    
    def getRects(self) -> list[Rect]:
        raise NotImplementedError()
    
    def getGrids(self) -> dict[Grid]:
         raise NotImplementedError()
    
    def removeRectGroup(self, label):
        raise NotImplementedError()
    
    def removeGrid(self, label):
        raise NotImplementedError()
    
    def appendGrid(self, text):
        raise NotImplementedError()
    
    def activateGrid(self, text):
        raise NotImplementedError()
    
    def getAIIgnoreRects(self) -> list:
        raise NotImplementedError()


from PyQt5.QtWidgets import QMainWindow, QPushButton
from PyQt5.QtGui import QStandardItem
from PyQt5.QtWidgets import QAction

from silicon_analyser.helper.abstract.abstractimage import AbstractImage
from silicon_analyser.helper.abstract.abstracttreehelper import AbstractTreeHelper

class AbstractMyWindow(QMainWindow):
    _actionGridAddRowTop: QAction
    _actionSaveModel: QAction
    _actionLoadModel: QAction
    _actionRemoveGrid: QAction
    _actionRemoveLabel: QAction
    _actionSaveAsCsv: QAction
    autosave: bool
    def __init__(self):
        QMainWindow.__init__(self)
        
    def getManualItem(self) -> QStandardItem:
        raise NotImplementedError()
    
    def getScale(self):
        raise NotImplementedError()
    
    def getPos(self):
        raise NotImplementedError()

    def getPosX(self):
        raise NotImplementedError()
    
    def getPosY(self):
        raise NotImplementedError()
    
    def setPosX(self, x):
        raise NotImplementedError()
    
    def setPosY(self, y):
        raise NotImplementedError()
    
    def appendGrid(self, text):
        raise NotImplementedError()
    
    def activateRectGroup(self, text):
        raise NotImplementedError()
    
    def deactivateRectGroup(self, text):
        raise NotImplementedError()
    
    def clearAIItem(self):
        raise NotImplementedError()
    
    def fetchData(self,x,y,ex,ey):
        raise NotImplementedError()

    def fetchFullData(self):
        raise NotImplementedError()
    
    def getAIIgnoreRects(self) -> list:
        raise NotImplementedError()
    
    def drawImage(self):
        raise NotImplementedError()
    
    def setStatusText(self, text):
        raise NotImplementedError()
    
    def imageWidth(self):
        raise NotImplementedError()

    def imageHeight(self):
        raise NotImplementedError()
    
    def reloadProperyWindowByGrid(self, grid):
        raise NotImplementedError()
    
    def reloadPropertyWindow(self):
        raise NotImplementedError()
    
    def getImage(self) -> AbstractImage:
        raise NotImplementedError()
    
    def getAIItem(self) -> QStandardItem:
        raise NotImplementedError()
        
    def getTree(self) -> AbstractTreeHelper:
        raise NotImplementedError()
    
    def setLastModel(self, name, model_train):
        raise NotImplementedError()
    
    def getModel(self, name):
        raise NotImplementedError()
    
    def hasModel(self, name) -> bool:
        raise NotImplementedError()
    
    def drawImgAndMinimap(self):
        raise NotImplementedError()



from PyQt5.QtCore import Qt, QRect, QSize, QTimer
from PyQt5.QtWidgets import QLabel, QSizePolicy
from PyQt5.QtGui import QImage, QPixmap, QColor, QMouseEvent, QPainter, QPen, QBrush
import numpy as np
from silicon_analyser.helper.abstract.abstractmywindow import AbstractMyWindow
from silicon_analyser.savefiles import saveGrids,saveRects
from silicon_analyser.grid import Grid
from silicon_analyser.rect import Rect
from silicon_analyser.treeitem import TreeItem

class FullImage(QLabel):
    _rects: dict[Rect]
    _aiRects: dict[Rect]
    _rectActive: dict[bool]
    _grids: dict[Grid]
    _aiGrids: dict[Grid]
    _gridsActive: dict[bool]
    _aiGridsActive: dict[bool]
    _moveStartX: int
    _moveStartY: int
    _moveDeltaX: int
    _moveDeltaY: int
    _moveTimer: QTimer
    
    def __init__(self, parent):
        QLabel.__init__(self, parent)
        self._drawRectStart = False
        self._rectStartX = 0
        self._rectStartY = 0
        self._rectEndX = 0
        self._rectEndY = 0
        self._grids = {}
        self._aiGrids = {}
        self._gridsActive = {}
        self._aiGridsActive = {}
        self._rects = {}
        self._aiRects = {}
        self._rectActive = {}
        self._aiRectActive = {}
        self._aiIgnoreRects = []
        self.setSizePolicy(QSizePolicy.Policy.Expanding,QSizePolicy.Policy.Expanding)
        self._moveTimer = QTimer()
        self._moveTimer.timeout.connect(self.moveUpdate)
        self._moveTimer.setInterval(int(1000 * 0.2))
        self._moveStart = False
    
    def moveUpdate(self):
        posX, posY = self._myWindow.getPos()
        posX += self._moveDeltaX
        posY += self._moveDeltaY
        self._myWindow.setPosX(posX)
        self._myWindow.setPosY(posY)
        self._myWindow.drawImgAndMinimap()
    
    def initialize(self, myWindow: AbstractMyWindow, pixmap: QPixmap):
        self._myWindow = myWindow
        self._pixmap: QPixmap = pixmap
        self._currentImg = pixmap
    
    def getRects(self) -> list[Rect]:
        return self._rects
    
    def getAiRects(self) -> list[Rect]:
        return self._aiRects
    
    def getAIIgnoreRects(self) -> list:
        return self._aiIgnoreRects
    
    def markGridItem(self,evx,evy,button):
        scale = self._myWindow.getScale()
        posX, posY = self._myWindow.getPos()
        sposX, sposY = posX*scale, posY*scale
        for k in self._grids.keys():
            if self._gridsActive[k]:
                grid: Grid = self._grids[k]
                x,y,ex,ey = grid.x, grid.y, grid.x + grid.width, grid.y + grid.height
                x = int(self._translatePixelToScaled(x)-sposX)
                y = int(self._translatePixelToScaled(y)-sposY)
                ex = int(self._translatePixelToScaled(ex)-sposX)
                ey = int(self._translatePixelToScaled(ey)-sposY)
                if ex<x:
                    x,ex = ex,x
                if ey<y:
                    y,ey = ey,y
                if x < evx and y < evy and ex > evx and ey > evy:
                    tevx = self._translateEventToPixel(evx)
                    tevy = self._translateEventToPixel(evy)
                    if button == Qt.LeftButton:
                        grid.setRect(int((tevx - (grid.x - posX))/(grid.width/grid.cols)),int((tevy - (grid.y - posY))/(grid.height/grid.rows)), self._myWindow.getTree().selectedLabel())
                    if button == Qt.RightButton:
                        grid.unsetRect(int((tevx - (grid.x - posX))/(grid.width/grid.cols)),int((tevy - (grid.y - posY))/(grid.height/grid.rows)), self._myWindow.getTree().selectedLabel())
                    if self._myWindow.autosave:
                        saveGrids(self._grids)
                    
    def mousePressEvent(self, event: QMouseEvent):
        if event.button() == Qt.MiddleButton:
            self._moveStart = True
            self._moveStartX = event.x()
            self._moveStartY = event.y()
            self._moveDeltaX = 0
            self._moveDeltaY = 0
        tree = self._myWindow.getTree()
        if tree.selectedType() == TreeItem.TYPE_GRID_ITEM:
            self.markGridItem(event.x(),event.y(),event.button())
            return
        if event.button() == Qt.LeftButton:
            print("FullImage: mousePressEvent",self._drawRectStart)
            if tree.selectedType() is not None:
                if not self._drawRectStart:
                    self._drawRectStart = True
                    self._rectStartX = self._translateEventToPixel(event.x())
                    self._rectStartY = self._translateEventToPixel(event.y())
    
    def mouseMoveEvent(self, event: QMouseEvent | None) -> None:
        if self._moveStart:
            self._moveDeltaX = event.x() - self._moveStartX
            self._moveDeltaY = event.y() - self._moveStartY
            self._moveTimer.start()

        if self._drawRectStart:
            self._rectEndX = self._translateEventToPixel(event.x())
            self._rectEndY = self._translateEventToPixel(event.y())
            if(self._rectEndX < self._rectStartX):
                self._rectEndX, self._rectStartX = self._rectStartX, self._rectEndX
            if(self._rectEndY < self._rectStartY):
                self._rectEndY, self._rectStartY = self._rectStartY, self._rectEndY
            pixmap = self._currentImg.copy()
            qp = QPainter(pixmap)
            brush = QBrush(QColor(255,0,0,127))
            pen = QPen(Qt.red, 2)
            qp.setBrush(brush)
            qp.setPen(pen)
            qp.drawRect(
                self._translatePixelToScaled(self._rectStartX),
                self._translatePixelToScaled(self._rectStartY),
                self._translatePixelToScaled(self._rectEndX)-self._translatePixelToScaled(self._rectStartX),
                self._translatePixelToScaled(self._rectEndY)-self._translatePixelToScaled(self._rectStartY))
            qp.end()
            self.setPixmap(pixmap)
     
    def mouseReleaseEvent(self, event: QMouseEvent):
        if event.button() == Qt.MiddleButton:
            self._moveStart = False
            self._moveTimer.stop()
            
        if event.button() == Qt.LeftButton:
            self._drawRectStart = False
            print("mouseReleaseEvent",self._rectStartX,self._rectStartY,self._rectEndX,self._rectEndY)
            selectedKey: str = self._myWindow.getTree().selectedLabel()
            if selectedKey is not None:
                x = self._rectStartX
                y = self._rectStartY
                ex = self._rectEndX
                ey = self._rectEndY
                if selectedKey in self._rects:
                    self.appendRect(selectedKey,x,y,ex,ey)
                if selectedKey in self._grids:
                    grid: Grid = self._grids[selectedKey]
                    posX, posY = self._myWindow.getPos()
                    grid.x = posX+x
                    grid.y = posY+y
                    grid.width = ex-x
                    grid.height = ey-y
                    self._grids[selectedKey] = grid
                    print(f"grid resized: {id(grid)}")
                    self._myWindow.reloadProperyWindowByGrid(grid)
                    if self._myWindow.autosave:
                        saveGrids(self._grids)
                self.drawImage()
        if event.button() == Qt.RightButton:
            print("right click")
            self.removeRectAt(event.x(),event.y())
            
    def removeRectGroup(self, label):
        del self._rects[label]
        del self._aiRects[label]
        del self._rectActive[label]
        del self._aiRectActive[label]
        
    def removeGrid(self, label):
        del self._grids[label]
        del self._aiGrids[label]
        del self._gridsActive[label]
        del self._aiGridsActive[label]

    def removeRectAt(self, evx, evy):
        scale = self._myWindow.getScale()
        posX, posY = self._myWindow.getPos()
        sposX, sposY = posX*scale, posY*scale
        for k in self._rects.keys():
            toRemove = []
            if self._rectActive[k]:
                for i in range(0,len(self._rects[k])):
                    rect: Rect = self._rects[k][i]
                    x,y,ex,ey = rect.x, rect.y, rect.ex, rect.ey
                    x = int(self._translatePixelToScaled(x)-sposX)
                    y = int(self._translatePixelToScaled(y)-sposY)
                    ex = int(self._translatePixelToScaled(ex)-sposX)
                    ey = int(self._translatePixelToScaled(ey)-sposY)
                        
                    if ex < x:
                        x,ex = ex,x
                    if ey < y:
                        y,ey = ey,y
                    if x < evx and y < evy and ex > evx and ey > evy:
                        toRemove.append(i)
            for i in  sorted(toRemove, reverse=True):
                print(f"remove {i}")
                del self._rects[k][i]
        
        for k in self._aiRects.keys():
            toRemove = []
            if self._aiRectActive[k]:
                for i in range(0,len(self._aiRects[k])):
                    rect: Rect = self._aiRects[k][i]
                    x,y,ex,ey = rect.x, rect.y, rect.ex, rect.ey
                    x = int(self._translatePixelToScaled(x)-sposX)
                    y = int(self._translatePixelToScaled(y)-sposY)
                    ex = int(self._translatePixelToScaled(ex)-sposX)
                    ey = int(self._translatePixelToScaled(ey)-sposY)
                        
                    if ex<x:
                        x,ex = ex,x
                    if ey<y:
                        y,ey = ey,y
                    if x<evx and y<evy and ex>evx and ey>evy:
                        toRemove.append(i)
            for i in sorted(toRemove, reverse=True):
                print(f"remove {i}")
                self._aiIgnoreRects.append(self._aiRects[k][i])
                del self._aiRects[k][i]
                
        self.drawImage()
        if  self._myWindow.autosave:
            saveRects(self._rects)
            saveGrids(self._grids)

    def appendRect(self,key,x,y,ex,ey):
        self._appendRect(key, self._rects, x, y, ex, ey)
        if self._myWindow.autosave:
            saveRects(self._rects)

    def _appendRect(self, key, rects, x, y, ex, ey, ignorePos = False):
        if ignorePos:
            rects[key].append(Rect(
                x,
                y,
                ex,
                ey
            ))
        else:
            posX, posY = self._myWindow.getPos()
            rects[key].append(Rect(
                posX+x,
                posY+y,
                posX+ex,
                posY+ey
            ))
                
    def appendAIRect(self,key,x,y,ex,ey):
        self._appendRect(key, self._aiRects, x, y, ex, ey, True)

    def fetchFullData(self):
        temp = QImage(self._pixmap.toImage())
        ptr = temp.constBits()
        ptr.setsize(temp.byteCount())
        arr = np.frombuffer(ptr, dtype=np.ubyte).reshape(temp.height(), temp.width(), 4)
        return arr
    
    def fetchData(self,x,y,ex,ey):
        x = int(x)
        ex = int(ex)
        y = int(y)
        ey = int(ey)
        arr = self.fetchFullData()
        r = arr[y:ey+1,x:ex+1,0:3]
        return r
        
    def loadRects(self, rects):
        self._rects = rects
        
    def loadGrids(self, grids):
        for g1_name in self._grids:
            for g2_name in grids:
                if(g1_name == g2_name):
                    self._grids[g1_name].replaceValues(grids[g2_name])
        
    def _translateEventToPixel(self, v):
        return int(v/self._myWindow.getScale())
    
    def _translatePixelToScaled(self, v):
        return int(v*self._myWindow.getScale())
    
    def clearAIRects(self):
        self._aiGridsActive = {}
        self._aiGrids = {}
        self._aiRects = {}
        self._aiRectActive = {}
    
    def drawRectOnScaledImg(self, scaledImg: QPixmap):
        qp = QPainter(scaledImg)
        brush = QBrush(QColor(0,0,255,80))
        pen = QPen(Qt.blue, 2)
        qp.setBrush(brush)
        qp.setPen(pen)
        self._drawRectOnScaledImg(self._rects, self._rectActive, qp)
        brush = QBrush(QColor(0,255,0,80))
        pen = QPen(Qt.green, 2)
        qp.setBrush(brush)
        qp.setPen(pen)
        self._drawRectOnScaledImg(self._aiRects, self._aiRectActive, qp)
        qp.end()
        
    def drawGridOnScaledImg(self, scaledImg: QPixmap):
        qp = QPainter(scaledImg)
        brush = QBrush(QColor(0,0,255,80))
        pen = QPen(Qt.blue, 2)
        qp.setBrush(brush)
        qp.setPen(pen)
        self._drawGridOnScaledImg(self._grids, self._gridsActive, qp, TreeItem.TYPE_GRID_ITEM)

        brush = QBrush(QColor(0,255,0,80))
        pen = QPen(QColor(0,0,0,0), 2)
        qp.setBrush(brush)
        qp.setPen(pen)
        self._drawGridOnScaledImg(self._aiGrids, self._aiGridsActive, qp, TreeItem.TYPE_AI_GRID_ITEM, QColor(0,255,0,40), QColor(0,255,0,127), QColor(0,0,0,0))

        qp.end()
    
    def calcGridCellsVisibleRange(self, grid):
        posX, posY = self._myWindow.getPos()
        scale = self._myWindow.getScale()
        cellWidth = grid.width / grid.cols
        cellHeight = grid.height / grid.rows
        
        startCol = -(grid.x - posX)//cellWidth
        startCol = int(max(0, startCol))
        
        startRow = -(grid.y - posY)//cellHeight
        startRow = int(max(0, startRow))
        
        endCol = int((grid.x + grid.width - posX)//cellWidth)
        endRow = int((grid.y + grid.height - posY)//cellHeight)
        
        maxColsCnt = (self.width() / cellWidth)/scale
        maxRowsCnt = (self.height() / cellHeight)/scale
        
        endCol = max(int(startCol + maxColsCnt),endCol)
        endRow = max(int(startRow + maxRowsCnt),endRow)
        endCol = int(min(endCol, grid.cols))
        endRow = int(min(endRow, grid.rows))
        
        return (startCol, startRow, endCol, endRow)
        
    def getGrids(self) -> dict[Grid]:
        return self._grids
        
    def _drawGridOnScaledImg(self, grids, gridsActive, qp:QPainter, gridItemType:str, rectSetColor:QColor = QColor(0,255,255,40), rectSetActiveColor:QColor = QColor(0,255,255,127), rectUnsetColor:QColor = QColor(0,0,255,20)):
        posX, posY = self._myWindow.getPos()
        scale = self._myWindow.getScale()
        sposX, sposY = posX*scale, posY*scale
        for k in grids.keys():
            if gridsActive[k]:
                grid: Grid = grids[k]
                cellWidth = grid.width / grid.cols
                cellHeight = grid.height / grid.rows
                startCol, startRow, endCol, endRow = self.calcGridCellsVisibleRange(grid)
                #for col in range(0,grid.cols):
                #    for row in range(0,grid.rows):
                for col in range(startCol,endCol):
                    for row in range(startRow,endRow):
                        ox = grid.x + col * cellWidth
                        oy = grid.y + row * cellHeight
                        ex = ox + cellWidth
                        ey = oy + cellHeight
                        x = int(self._translatePixelToScaled(ox)-sposX)
                        y = int(self._translatePixelToScaled(oy)-sposY)
                        ex = int(self._translatePixelToScaled(ex)-sposX)
                        ey = int(self._translatePixelToScaled(ey)-sposY)
                        w = ex - x
                        h = ey - y
                        if x>=0 and y>=0 and x<=self.width() and y<=self.height():
                            if grid.isRectSet(col,row):
                                rectLabel = grid.rectLabel(col, row)
                                if self._myWindow.getTree().isItemSelected(rectLabel, grid.name, gridItemType):
                                    brush = QBrush(rectSetActiveColor)
                                else:
                                    brush = QBrush(rectSetColor)
                                qp.setBrush(brush)
                            else:
                                brush = QBrush(rectUnsetColor)
                                qp.setBrush(brush)
                            qp.drawRect(x,y,w,h)
                        
    def _drawRectOnScaledImg(self, rects, rectActive, qp:QPainter):
        posX, posY = self._myWindow.getPos()
        scale = self._myWindow.getScale()
        sposX, sposY = posX*scale, posY*scale
        for k in rects.keys():
            if rectActive[k]:
                for r in rects[k]:
                    x,y,ex,ey = r.x, r.y, r.ex, r.ey
                    x = int(self._translatePixelToScaled(x)-sposX)
                    y = int(self._translatePixelToScaled(y)-sposY)
                    ex = int(self._translatePixelToScaled(ex)-sposX)
                    ey = int(self._translatePixelToScaled(ey)-sposY)
                    w = ex - x
                    h = ey - y
                    if x>=0 and y>=0 and x<=self.width() and y<=self.height():
                        qp.drawRect(x,y,w,h)

    def drawImage(self):
        posX, posY = self._myWindow.getPos()
        scale = self._myWindow.getScale()
        rect = QRect(int(posX), int(posY), int(self.width()/scale), int(self.height()/scale))
        cropped = self._pixmap.copy(rect)
        scaled = cropped.scaled(QSize(int(self.width()), int(self.height())))
        self._currentImg = scaled
        self.drawRectOnScaledImg(scaled)
        self.drawGridOnScaledImg(scaled)
        self.setPixmap(scaled)
    
    def appendAIGrid(self, text):
        grid = self._grids[text]
        aiGrid = Grid(grid.name,grid.x,grid.y,grid.cols,grid.rows,grid.width,grid.height)
        self._aiGrids[text] = aiGrid
        return aiGrid
    
    def appendGrid(self, text):
        grid = Grid(text,0,0,20,20,20*10,20*10)
        self._grids[text] = grid
        self._myWindow.reloadProperyWindowByGrid(grid)
        if self._myWindow.autosave:
            saveGrids(self._grids)
        return self._grids[text]
    
    def activateAIGrid(self, text):
        self._aiGridsActive[text] = True
        self.drawImage()

    def activateGrid(self, text):
        self._gridsActive[text] = True
        self.drawImage()

    def deactivateGrid(self, text):
        self._gridsActive[text] = False
        self.drawImage()
    
    def appendAIGridRectGroup(self, grid: Grid, text):
        grid.addRectGroup(text)
    
    def appendGridRectGroup(self, grid: Grid, text):
        grid.addRectGroup(text)
    
    def activateAIGridRectGroup(self, grid: Grid, text):
        grid.rectActive(text)

    def deactivateAIGridRectGroup(self, grid: Grid, text):
        grid.rectDeactive(text)

    def activateGridRectGroup(self, grid: Grid, text):
        grid.rectActive(text)

    def deactivateGridRectGroup(self, grid: Grid, text):
        grid.rectDeactive(text)

    def appendRectGroup(self, text):
        self._rects[text] = []
    
    def activateRectGroup(self, text):
        self._rectActive[text] = True
        self.drawImage()

    def deactivateRectGroup(self, text):
        self._rectActive[text] = False
        self.drawImage()
        
    def appendAIRectGroup(self, text):
        self._aiRects[text] = []
    
    def activateAIRectGroup(self, text):
        self._aiRectActive[text] = True
        self.drawImage()

    def deactivateAIRectGroup(self, text):
        self._aiRectActive[text] = False
        self.drawImage()



class Grid:
    _rects: dict[list[list[int]]]
    _rectsActive: dict[bool]
    def __init__(self, name, x, y, cols, rows, width, height):
        self.name = name
        self.x = x
        self.y = y
        self.cols = cols
        self.rows = rows
        self.width = width
        self.height = height
        self._rects = {}
        self._rectsActive = {}
        print(f"init grid:{id(self)}")
    
    def replaceValues(self,grid):
        self.x = grid.x
        self.y = grid.y
        self.cols = grid.cols
        self.rows = grid.rows
        self.width = grid.width
        self.height = grid.height
        self._rects = grid._rects
        for k in self._rects:
            for cx,cy in list(self._rects[k]):
                if cx < 0 or cy < 0:
                    self._rects[k].remove([cx,cy])
        self._rectsActive = grid._rectsActive
        
    def getLabels(self) -> list[str]:
        return self._rects.keys()
    
    def removeRectGroup(self, label):
        del self._rects[label]
        del self._rectsActive[label]
    
    def addRectGroup(self, text):
        self._rects[text] = []
        self.rectActive(text)
    
    def rectActive(self, text):
        self._rectsActive[text] = True
    
    def rectDeactive(self, text):
        self._rectsActive[text] = False
        
    def addTopRow(self):
        self.y -= int(self.height/self.rows)
        self.rows += 1
        for label in self._rects:
            for i in range(0,len(self._rects[label])):
                col, row = self._rects[label][i]
                self._rects[label][i] = [col, row+1]
    
    def setRect(self, col, row, label):
        #print(f"setRect {col} {row} {label}")
        r = [col, row]
        if label not in self._rects:
            self._rects[label] = []
        if r in self._rects[label]:
            self._rects[label].remove(r)
        self._rects[label].append(r)
        
    def unsetRect(self, col, row, label):
        print(f"unsetRect {col} {row}")
        r = [col, row]
        if r in self._rects[label]:
            self._rects[label].remove(r)
    
    def rectLabel(self, col, row) -> str:
        r = [col, row]
        keys = self._rects.keys()
        for k in keys:
            if self._rectsActive[k]:
                if r in self._rects[k]:
                    return k
    
    def isRectSet(self, col ,row, key = None) -> bool:
        r = [col, row]
        if key is None:
            keys = self._rects.keys()
        else:
            keys = [key]
        for k in keys:
            if self._rectsActive[k]:
                if r in self._rects[k]:
                    return True
        return False
    
    
def getAllCellRects(grid: Grid):
    cellRects = []
    for cx in range(0,grid.cols):
        for cy in range(0,grid.rows):
            cellRects.append((cx,cy))
    return cellRects


from PyQt5 import QtCore, uic
from PyQt5.QtCore import Qt, QItemSelection
from PyQt5.QtWidgets import QFileDialog, QTableView, QStatusBar, QAction, QPushButton
from PyQt5.QtGui import QPixmap, QStandardItem, QStandardItemModel
from os import path as p
import sys

import silicon_analyser.savefiles
from silicon_analyser.savefiles import loadGrids, loadRects
from silicon_analyser.helper.abstract.abstracttreehelper import AbstractTreeHelper
from silicon_analyser.helper.abstract.abstractimage import AbstractImage
from silicon_analyser.helper.abstract.abstractmywindow import AbstractMyWindow
from silicon_analyser.helper.minimap import MiniMap
from silicon_analyser.helper.fullimage import FullImage
from silicon_analyser.helper.addlabelbtn import AddLabelBtn
from silicon_analyser.helper.computebtn import ComputeBtn
from silicon_analyser.helper.addgridbtn import AddGridBtn
from silicon_analyser.treeitem import TreeItem
from silicon_analyser.helper.properties import PropertiesUtil
from silicon_analyser.grid import Grid
from silicon_analyser.treeitem import TreeItem
from silicon_analyser.helper.tree import Tree

class MyWindow(AbstractMyWindow):
    _tree: Tree
    _properties: QTableView
    _addLabelBtn: AddLabelBtn
    _computeBtn: ComputeBtn
    _addGridBtn: AddGridBtn
    _statusBar: QStatusBar
    _minimap: MiniMap
    _image: FullImage
    _models: dict
    _actionGridAddRowTop: QAction
    _actionSaveModel: QAction
    _actionLoadModel: QAction
    _actionRemoveGrid: QAction
    _actionRemoveLabel: QAction
    _actionMade_by_TheCrazyT: QAction
    _actionUrl: QAction
    _actionSaveAsCsv: QAction
    autosave: bool
    
    def __init__(self):
        AbstractMyWindow.__init__(self)
        uic.loadUi(p.abspath(p.join(p.dirname(__file__), '.')) + '/main_window.ui', self)
        tree: Tree = self._tree
        properties: QTableView = self._properties
                
        self._treeModel = QStandardItemModel()
        self._propertiesModel = QStandardItemModel()
        properties.setModel(self._propertiesModel)
        self._propertiesUtil = PropertiesUtil(self,self._properties)
        
        self._treeManualItem = QStandardItem("Manual")
        self._treeManualItem.setFlags(QtCore.Qt.ItemIsUserCheckable |
                          QtCore.Qt.ItemIsEnabled)
        self._treeManualItem.setCheckState(QtCore.Qt.Unchecked)
        self._treeModel.appendRow(self._treeManualItem)
        self._treeAIItem = QStandardItem("AI")
        self._treeAIItem.setFlags(QtCore.Qt.ItemIsUserCheckable |
                          QtCore.Qt.ItemIsEnabled)
        self._treeAIItem.setCheckState(QtCore.Qt.Unchecked)
        self._treeModel.appendRow(self._treeAIItem)
        tree.setModel(self._treeModel)
        self._models = {}
        self._actionUrl.triggered.connect(self.openMainUrl)

        if(len(sys.argv) < 2):
            dlg = QFileDialog()
            filenames = dlg.getOpenFileName(caption="Load image",filter="Image (*.png;*.jpg;*.bmp;*.gif)",initialFilter="Trained model (*.h5)")
            if(len(filenames) >= 1):
                self._pixmap = QPixmap(filenames[0])
        else:
            self._pixmap = QPixmap(sys.argv[1])
        
        computeBtn: ComputeBtn = self._computeBtn
        addGridBtn: AddGridBtn = self._addGridBtn
        addLabelBtn: AddLabelBtn = self._addLabelBtn
        minimap: MiniMap = self._minimap
        image: FullImage = self._image

        tree.initialize(self)
        computeBtn.initialize(self)
        addGridBtn.initialize(self)
        addLabelBtn.initialize(self)
        minimap.initialize(self, self._pixmap)
        image.initialize(self, self._pixmap)
        
        computeBtn.setDisabled(True)
        addLabelBtn.setDisabled(True)
        tree.evtTreeSelectionChanged.connect(self.treeSelectionChanged)
        
        self._posX = 0
        self._posY = 0
        
        self._scale = 1.0
        self.drawImgAndMinimap()
        
        self.autosave = False
        
        if p.isfile(silicon_analyser.savefiles.SAVE_RECTS):
            with open(silicon_analyser.savefiles.SAVE_RECTS,"r") as f:
                rects = loadRects()
                for k in rects.keys():
                    self.getTree().addTreeItem(k)
                image.loadRects(rects)
        if p.isfile(silicon_analyser.savefiles.SAVE_GRIDS):
            grids: dict[Grid] = loadGrids()
            for gridKey in grids.keys():
                grid, parentTreeItem = self.getTree().addTreeItem(gridKey,TreeItem.TYPE_GRID)
                for gridItemKey in grids[gridKey].getLabels():
                    _, treeItem = self.getTree().addTreeItem(gridItemKey,TreeItem.TYPE_GRID_ITEM, grid, parentTreeItem)
                    text = treeItem.data(TreeItem.TEXT)
                    if(not grids[gridKey]._rectsActive[text]):
                        treeItem.setCheckState(QtCore.Qt.Unchecked)
            image.loadGrids(grids)
        
        image.drawImage()
        self.autosave = True
        
    def treeSelectionChanged(self, selection: QItemSelection):
        tree: Tree = self._tree
        selectedType: str = tree.selectedType()
        computeBtn: ComputeBtn = self._computeBtn
        addLabelBtn: AddLabelBtn = self._addLabelBtn
        if((selectedType == TreeItem.TYPE_GRID_ITEM)
           or (selectedType == TreeItem.TYPE_GRID)
           or (selectedType == TreeItem.TYPE_AI_GRID)
           or (selectedType == TreeItem.TYPE_AI_GRID_ITEM)
           ):
            computeBtn.setDisabled(False)
            addLabelBtn.setDisabled(False)
        else:
            computeBtn.setDisabled(True)
            addLabelBtn.setDisabled(True)

    def openMainUrl(self):
        import webbrowser
        return webbrowser.open('https://github.com/TheCrazyT/SiliconAnalyser')
        
    def getModel(self, name):
        if name in self._models:
            return self._models[name]
        return None
    
    def hasModel(self, name) -> bool:
        if name in self._models:
            return True
        return False
    
    def setLastModel(self, name, model_train):
        self._models[name] = model_train
    
    def getImage(self) -> AbstractImage:
        return self._image
    
    def getMinimap(self) -> MiniMap:
        return self._minimap

    def imageWidth(self):
        return self.getImage().width()

    def imageHeight(self):
        return self.getImage().height()
            
    def setStatusText(self, text):
        self._statusBar.showMessage(text)
        
    def getTree(self) -> AbstractTreeHelper:
        return self._tree
    
    def getManualItem(self) -> QStandardItem:
        return self._treeManualItem
    
    def getAIItem(self) -> QStandardItem:
        return self._treeAIItem
    
    def getScale(self):
        return self._scale
    
    def getPos(self):
        return self._posX, self._posY

    def getPosX(self):
        return self._posX
    
    def getPosY(self):
        return self._posY
    
    def setPosX(self, x):
        self._posX = x
    
    def setPosY(self, y):
        self._posY = y
    
    def drawImgAndMinimap(self):
        self.getImage().drawImage()
        self.getMinimap().drawMinimap()
    
    def reloadProperyWindowByGrid(self, grid):
        return self._propertiesUtil.reloadProperyWindowByGrid(grid)
    
    def reloadPropertyWindow(self, selection: QItemSelection):
        return self._propertiesUtil.reloadPropertyWindow(selection)
    
    def keyPressEvent(self, event):
        c = 10
        if event.modifiers() & Qt.ShiftModifier:
            c = 100
        if event.key() == Qt.Key_Left:
            self._posX -= c
            self._posX = max(self._posX,0)
        if event.key() == Qt.Key_Right:
            self._posX += c
            self._posX = min(self._posX,self._pixmap.width())
        if event.key() == Qt.Key_Up:
            self._posY -= c
            self._posY = max(self._posY,0)
        if event.key() == Qt.Key_Down:
            self._posY += c
            self._posY = min(self._posY,self._pixmap.height())
        self.drawImgAndMinimap()
            
    def wheelEvent(self,event):
        #print(event.angleDelta())
        if event.angleDelta().y() > 0:
            self._scale += 0.25
        else:
            self._scale -= 0.25
        self._scale = max(self._scale,0.25)
        print(self._scale)
        self.drawImgAndMinimap()


import json
from json import JSONDecoder, JSONEncoder

from silicon_analyser.grid import Grid
from silicon_analyser.rect import Rect

SAVE_RECTS = "rects.json"
SAVE_GRIDS = "grids.json"
doSaveRects = False
doSaveGrids = False

class JSONGridDecoder(JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, *args, **kwargs)
    def decode(self, *args, **kwargs):
        d = super().decode(*args, **kwargs)
        grids = {}
        for gridName in d:
            item = d[gridName]
            name = item["name"]
            if gridName != name:
                raise Exception("invalid gridname")
            x = item["x"]
            y = item["y"]
            cols = item["cols"]
            rows = item["rows"]
            width = item["width"]
            height = item["height"]
            g: Grid = Grid(name, x , y, cols, rows, width, height)
            g._rects = item["_rects"]
            g._rectsActive = item["_rectsActive"]
            grids[gridName] = g
        return grids
        
class MyJSONEncoder(JSONEncoder):
        def default(self, o):
            return o.__dict__

def loadRects() -> dict[Rect]:
    with open(SAVE_RECTS,"r") as f:
        return json.load(f)

def loadGrids() -> dict[Grid]:
    with open(SAVE_GRIDS,"r") as f:
        return json.load(f, cls=JSONGridDecoder)
    
def triggerSaveRects():
    global doSaveRects
    print("triggerSaveRects")
    doSaveRects = True

def triggerSaveGrids():
    global doSaveGrids
    print("triggerSaveGrids")
    doSaveGrids = True

def saveRects(rects):
    global doSaveRects
    print("saveRects")
    if doSaveRects:
        with open(SAVE_RECTS,"w") as f:
            json.dump(rects, f, cls = MyJSONEncoder)
        doSaveRects = False

def saveGrids(grids):
    global doSaveGrids
    print("saveGrids")
    if doSaveGrids:
        with open(SAVE_GRIDS,"w") as f:
            json.dump(grids, f, cls = MyJSONEncoder)
        doSaveGrids = False